# External Imports
import string
from enum import Enum, StrEnum, auto

import pandas as pd
from frozendict import frozendict


class PrimitiveDataTypes(StrEnum):
    """
    Enum representing primitive data types.
    """

    BOOLEAN = auto()
    FLOAT = auto()
    INTEGER = auto()
    STRING = auto()


class AbstractDataTypes(StrEnum):
    """
    Enum representing abstract data types.
    """

    TEXT = auto()
    DECIMAL = auto()
    EMAIL = auto()


class Constants(StrEnum):
    """
    Enum representing various constants.
    """

    STANDARD_COLUMN = auto()
    DATA_TYPE = auto()
    DATE_ORDER_TUPLE = auto()
    TRANSFORMATIONS = auto()
    VALIDATIONS = auto()
    POST_PROCESSING_DATA_TYPE = auto()
    COMPARISON_OPERATIONS = auto()
    IS_OPTIONAL = auto()
    COALESCE_CHILD_REQUIRED = auto()


class ComparisonOperations(StrEnum):
    """
    Enum representing comparison operations and their corresponding operators.
    """

    LESS_THAN = auto()
    LESS_THAN_OPERATOR = "<"
    LESS_THAN_EQUALS = auto()
    LESS_THAN_EQUALS_OPERATOR = "<="
    GREATER_THAN = auto()
    GREATER_THAN_OPERATOR = ">"
    GREATER_THAN_EQUALS = auto()
    GREATER_THAN_EQUALS_OPERATOR = ">="
    EQUALS = auto()
    EQUALS_OPERATOR = "=="
    NOT_EQUALS = auto()
    NOT_EQUALS_OPERATOR = "!="


class CalendarDateComponents(StrEnum):
    """
    Enum representing calendar date components.
    """

    DAY = auto()
    MONTH = auto()
    YEAR = auto()


class DateOrderTuples(Enum):
    """
    Enum representing different date orders.
    """

    DAY_MONTH_YEAR = (
        CalendarDateComponents.DAY.value,
        CalendarDateComponents.MONTH.value,
        CalendarDateComponents.YEAR.value,
    )
    MONTH_DAY_YEAR = (
        CalendarDateComponents.MONTH.value,
        CalendarDateComponents.DAY.value,
        CalendarDateComponents.YEAR.value,
    )
    YEAR_DAY_MONTH = (
        CalendarDateComponents.YEAR.value,
        CalendarDateComponents.DAY.value,
        CalendarDateComponents.MONTH.value,
    )
    YEAR_MONTH_DAY = (
        CalendarDateComponents.YEAR.value,
        CalendarDateComponents.MONTH.value,
        CalendarDateComponents.DAY.value,
    )


class DefaultColumns(StrEnum):
    """
    Enum representing default column names.
    """

    VALUE_PLACEHOLDER = "Value_Placeholder"
    ERROR = "Error"
    IS_ALLOWED = "Is_Allowed"
    REASON = "Reason"


# Create a translation table with all punctuation characters mapped to ''
STRING_TRANSLATION_TABLE = str.maketrans("", "", string.punctuation)

# Create a translation table with all digits characters mapped to ''
DIGIT_TRANSLATION_TABLE = str.maketrans("", "", string.digits)

PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP = frozendict(
    {
        int: pd.Int64Dtype(),
        float: pd.Float64Dtype(),
    }
)



# External Imports
import re
from datetime import datetime

import phonenumbers
from dateutil import parser
from loguru import logger
from phonenumbers import is_valid_number, parse

# Internal Imports
from sanctify.constants import DIGIT_TRANSLATION_TABLE, STRING_TRANSLATION_TABLE, CalendarDateComponents
from sanctify.exception import ValidationError


class Transformer:
    @staticmethod
    def parse_boolean(value: str):
        """
        Parses a boolean value from a string.

        Args:
            value (str): The input string.

        Returns:
            bool: The parsed boolean value.

        Raises:
            ValidationError: If the input string is not a valid boolean.
        """
        bool_string = str(value).lower()
        if bool_string in {"true", "false"}:
            return str(value).lower() == "true"
        else:
            raise ValidationError(f"Invalid boolean input: {bool_string = }")

    @staticmethod
    def remove_dot_from_string(value: str) -> str:
        """
        Removes all dots from a string except for "jr." and "sr.".

        Args:
            value (str): The input string.

        Returns:
            str: The string with dots removed.
        """
        return str(value).replace(".", "")

    @staticmethod
    def remove_punctuations(value: str) -> str:
        """
        Removes all punctuations from a string.

        Args:
            value (str): The input string.

        Returns:
            str: The string with punctuations removed.
        """
        # Remove punctuation using the translation table
        return str(value).translate(STRING_TRANSLATION_TABLE)

    @staticmethod
    def remove_all_digits(value: str) -> str:
        """
        Removes all digits from a string.

        Args:
            value (str): The input string.

        Returns:
            str: The string with digits removed.
        """
        # Remove digits using the translation table
        return str(value).translate(DIGIT_TRANSLATION_TABLE)

    @staticmethod
    def remove_punctuations_and_leading_zeroes_except_plus(value: str) -> str:
        """
        Removes all characters from a string except for digits and the plus sign.
        Strips leading zeroes from the resulting string, while preserving the plus sign if it exists.

        Args:
            value (str): The input string from which punctuations and leading zeroes need to be removed.

        Returns:
            str: The resulting string after removing punctuations and leading zeroes, with the plus sign preserved if it exists.
        """
        # Remove all characters except digits and plus sign, then strip leading zeroes
        cleaned_value = re.sub(r"[^\d+]", "", str(value))
        # Check if the string starts with a plus sign
        if cleaned_value.startswith("+"):
            # Preserve the plus sign, remove leading zeroes after the plus sign
            return "+" + cleaned_value[1:].lstrip("0")
        else:
            # If there's no plus sign, simply remove leading zeroes
            return cleaned_value.lstrip("0")

    @staticmethod
    def remove_all_spaces(value: str) -> str:
        """
        Removes all spaces from a string.

        Args:
            value (str): The input string.

        Returns:
            str: The string with spaces removed.
        """
        return re.sub(r"\s", "", str(value))

    @staticmethod
    def convert_to_lowercase(value: str) -> str:
        """
        Converts a string to lowercase.

        Args:
            value (str): The input string.

        Returns:
            str: The lowercase string.
        """
        return str(value).lower()

    @staticmethod
    def convert_to_uppercase(value: str) -> str:
        """
        Converts a string to uppercase.

        Args:
            value (str): The input string.

        Returns:
            str: The uppercase string.
        """
        return str(value).upper()

    @staticmethod
    def convert_to_titlecase(value: str) -> str:
        """
        Converts a string to titlecase.

        Args:
            value (str): The input string.

        Returns:
            str: The titlecase string.
        """
        return str(value).title()

    @staticmethod
    def replace_ii_with_II(value: str) -> str:
        """
        Replaces "ii","iI","Ii","II" with "II" in a string.

        Args:
            value (str): The input string.

        Returns:
            str: The string with "ii" replaced by "II".
        """
        return re.sub(r"\b[iI]{2}\b", "II", str(value))

    @staticmethod
    def convert_jr_to_Junior(value: str) -> str:
        """
        Replaces "jr., Jr." with "Junior." in a string.

        Args:
            value (str): The input string.

        Returns:
            str: The string with "jr." replaced by "Junior.".
        """
        return re.sub(r"\b[Jj]r\.", "Junior ", str(value))

    @staticmethod
    def convert_sr_to_Senior(value: str) -> str:
        """
        Replaces "sr., Sr." with "Senior." in a string.

        Args:
            value (str): The input string.

        Returns:
            str: The string with "sr." replaced by "Senior.".
        """
        return re.sub(r"\b[Ss]r\.", "Senior ", str(value))

    @staticmethod
    def parse_date_from_string(
        value: str, date_order_tuple: tuple, return_datetime: bool = False
    ) -> str | datetime | NotImplementedError | parser.ParserError:
        """
        Parses a date from a string using a specified date order.

        Args:
            value (str): The input string representing the date.
            date_order_tuple (tuple): The tuple specifying the order of date components (day, month, year).
            return_datetime (bool, optional): Whether to return a datetime object instead of a string date.
                Defaults to False.

        Returns:
            str | datetime: The parsed date as a string or datetime object.

        Raises:
            NotImplementedError: If the date order tuple is invalid.
            parser.ParserError: If the input string cannot be parsed as a date.
        """
        logger.debug(f"In parse_date_from_string received: {value = } | {date_order_tuple = }")
        error_message = f"({type(value)}, {value}) | {date_order_tuple = } | {return_datetime = }"

        try:
            value = str(value)
            day_first = date_order_tuple.index(CalendarDateComponents.DAY.value) < date_order_tuple.index(
                CalendarDateComponents.MONTH.value
            )
            year_first = date_order_tuple.index(CalendarDateComponents.YEAR.value) < date_order_tuple.index(
                CalendarDateComponents.MONTH.value
            )
        except ValueError as value_err:
            err = NotImplementedError(f"Invalid `date_order` tuple | {date_order_tuple = } | {value_err = }")
            logger.exception(f"{str(err) = }")
            raise err

        try:
            parsed_datetime = parser.parse(value, dayfirst=day_first, yearfirst=year_first)
        except parser.ParserError as err:
            logger.exception(f"{str(err) = } | DEBUG: {error_message}")
            raise ValidationError(f"Invalid date passed: {str(err)}")
        else:
            if return_datetime is True:
                return parsed_datetime
            else:
                return str(parsed_datetime.date())

    @staticmethod
    def extract_digits_from_string(value: str):
        """
        Extracts only the digits from a string.

        Args:
            value (str): The input string.

        Returns:
            str: The string containing only the digits.
        """
        return re.sub(r"\D", "", str(value))

    @staticmethod
    def extract_currency_from_amount(value: str) -> str:
        """
        Extracts the currency symbol from an amount string.

        Args:
            value (str): The input amount string.

        Returns:
            str: The extracted currency symbol.
        """
        match = re.search(r"[^\d.,\s]+", str(value))
        if match:
            return match.group()
        else:
            return ""

    @staticmethod
    def remove_currency_from_amount(value: str) -> str | ValidationError:
        """
        Removes currency symbols and separators from an amount string.

        Args:
            value (str): The input amount string.

        Returns:
            str: The amount string with currency symbols and separators removed.
            ValidationError: if amount is invalid or negative
        """
        # Use regex to remove everything except digits, decimal point, and minus sign
        _amount = re.sub(r"[^\d.-]", "", str(value))

        try:
            amount = float(_amount)
            if amount > 0:
                return str(amount)
            else:
                raise ValidationError(f"Amount {str(amount)} should not be negative")

        except ValueError:
            raise ValidationError(f"Failed to parse amount {str(_amount)}")

    @staticmethod
    def transform_str_into_phone_number_obj(
        value: str, region: str | None = "US"
    ) -> phonenumbers.PhoneNumber | ValidationError:
        """
        Transforms a phone number string into phone number object.

        Args:
            value (str): The phone number string from which the phone number object needs to be extracted.
            region (str | default = "US"): Here are some examples of valid country codes you can use as the region parameter:
                'US': United States
                'GB': United Kingdom
                'CA': Canada
                'AU': Australia
                'JP': Japan
                'DE': Germany
                You can find a comprehensive list of country codes in the ISO 3166-1 standard documentation.

        Returns:
            phonenumbers.PhoneNumber | ValidationError: The extracted phone number object, or a ValidationError if the phone number is invalid.

        Raises:
            ValidationError: If the phone number is invalid or cannot be parsed.
        """
        try:
            err_msg = f"Invalid/Fake phone_number = '{value}'"
            cleaned_number = Transformer.remove_punctuations_and_leading_zeroes_except_plus(value=value)
            parsed_number = parse(number=cleaned_number, region=region)
            if not is_valid_number(numobj=parsed_number):
                raise ValidationError(err_msg)

            return parsed_number

        except phonenumbers.NumberParseException:
            # Raise an error if the phone number is invalid
            raise ValidationError(err_msg)

        except Exception as err:
            raise ValidationError(f"Failed to Parse Phone Number '{value}' | {str(err)}")

    @staticmethod
    def extract_phone_number(value: str, region: str | None = "US") -> str | ValidationError:
        """
        Extracts the national phone number from a phone number string.

        Args:
            value (str): The phone number string from which the national phone number needs to be extracted.
            region (str | default = "US"): Here are some examples of valid country codes you can use as the region parameter:
                'US': United States
                'GB': United Kingdom
                'CA': Canada
                'AU': Australia
                'JP': Japan
                'DE': Germany
                You can find a comprehensive list of country codes in the ISO 3166-1 standard documentation.

        Returns:
            str | ValidationError: The extracted national phone number as a string, or a ValidationError if the phone number is invalid.
        """
        parsed_number: phonenumbers.PhoneNumber = Transformer.transform_str_into_phone_number_obj(
            value=value, region=region
        )
        return str(parsed_number.national_number)

    @staticmethod
    def extract_country_code(value: str, region: str | None = "US") -> str | ValidationError:
        """
        Extracts the country code from a phone number string.

        Args:
            value (str): The phone number string from which the country code needs to be extracted.
            region (str | default = "US"): Here are some examples of valid country codes you can use as the region parameter:
                'US': United States
                'GB': United Kingdom
                'CA': Canada
                'AU': Australia
                'JP': Japan
                'DE': Germany
                You can find a comprehensive list of country codes in the ISO 3166-1 standard documentation.

        Returns:
            str | ValidationError: The extracted country code as a string, or a ValidationError if the phone number is invalid.
        """
        parsed_number: phonenumbers.PhoneNumber = Transformer.transform_str_into_phone_number_obj(
            value=value, region=region
        )
        return str(parsed_number.country_code)



# External Imports
import ast
from datetime import date

from dateutil.relativedelta import relativedelta
from loguru import logger

# Internal Imports
from sanctify.constants import ComparisonOperations

COMPARISON_OPERATIONS_VS_FUNCTIONS_MAP = {
    ComparisonOperations.LESS_THAN.value: lambda x, y: x < y,
    ComparisonOperations.LESS_THAN_OPERATOR.value: lambda x, y: x < y,
    ComparisonOperations.LESS_THAN_EQUALS.value: lambda x, y: x <= y,
    ComparisonOperations.LESS_THAN_EQUALS_OPERATOR.value: lambda x, y: x <= y,
    ComparisonOperations.GREATER_THAN.value: lambda x, y: x > y,
    ComparisonOperations.GREATER_THAN_OPERATOR.value: lambda x, y: x > y,
    ComparisonOperations.GREATER_THAN_EQUALS.value: lambda x, y: x >= y,
    ComparisonOperations.GREATER_THAN_EQUALS_OPERATOR.value: lambda x, y: x >= y,
    ComparisonOperations.EQUALS.value: lambda x, y: x == y,
    ComparisonOperations.EQUALS_OPERATOR.value: lambda x, y: x == y,
    ComparisonOperations.NOT_EQUALS.value: lambda x, y: x != y,
    ComparisonOperations.NOT_EQUALS_OPERATOR.value: lambda x, y: x != y,
}


def replace_dict_keys(original_dict: dict, key_mapping: dict) -> dict:
    """
    Replace keys in a dictionary with the corresponding values from a key mapping dictionary.

    Args:
        original_dict (dict): Original dictionary.
        key_mapping (dict): Dictionary mapping old keys to new keys.

    Returns:
        dict: Dictionary with replaced keys.
    """
    return {key_mapping.get(key, key): value for key, value in original_dict.items()}


def match_comparision_operation(
    value, operation: str, operand: str | int
) -> bool | NotImplementedError | Exception | ValueError:
    """
    Perform a comparison operation between a value and an operand based on the specified operation.

    Args:
        value: Value to compare.
        operation (str): Comparison operation.
        operand (str|int): Operand for comparison.

    Returns:
        bool: Result of the comparison.

    Raises:
        NotImplementedError: If the operation is not supported.
        TypeError: If the comparison fails due to a type mismatch.
        Exception: If an unexpected exception occurs.
        ValueError: If the operation or operand is invalid.
    """
    err_message = f"{operation = } with {operand = } | DEBUG: {type(value) = }, {value = }"
    try:
        # Perform the comparison operation based on the provided operation

        if operation in COMPARISON_OPERATIONS_VS_FUNCTIONS_MAP:
            return COMPARISON_OPERATIONS_VS_FUNCTIONS_MAP[operation](value, operand)
        else:
            raise NotImplementedError(f"Unsupported Comparision: {err_message}")

    except TypeError as err:
        err = TypeError(f"System Error: {err_message}")
        logger.error(f"{str(err) = }")
        raise err


def calculate_age(date_of_birth: date, current_date: date = date.today()) -> int:
    """
    Calculate the age based on the date of birth.

    Args:
        date_of_birth (date): Date of birth.

    Returns:
        int: Calculated age.

    Notes:
        The age is calculated by subtracting the birth year from the current year,
        and adjusting for the birth month and day.

    Example:
        If the date of birth is '1990-05-15', and the current date is '2022-08-10',
        the calculated age will be 32.
    """
    return relativedelta(current_date, date_of_birth).years


def convert_tuples_to_lists(input_data):
    """
    Recursively converts tuples to lists in a nested dictionary.

    This function takes a nested data structure (dictionary, list, or tuple) as input
    and traverses through it recursively to find tuples. Whenever a tuple is found,
    it converts it into a list. The function handles nested dictionaries, lists, and tuples.

    Parameters:
        input_data (dict, list, tuple, or any): The nested data structure to process.

    Returns:
        The input data structure with tuples converted to lists.
    """
    if isinstance(input_data, dict):
        # If the input is a dictionary, process its key-value pairs recursively.
        return {key: convert_tuples_to_lists(value) for key, value in input_data.items()}
    elif isinstance(input_data, (tuple, list)):
        # If the input is a tuple or list, process its elements recursively.
        return [convert_tuples_to_lists(item) for item in input_data]
    else:
        # If the input is not a dictionary, tuple, or list, return it unchanged.
        return input_data


@staticmethod
def convert_stringified_list_to_list(value: str) -> list | str:
    """
    Converts a stringified list to a list.

    Args:
        value (str): The input string representing a list.

    Returns:
        list | str: The converted list if successful, otherwise the original string.

    """
    try:
        return ast.literal_eval(value)
    except (SyntaxError, ValueError):
        return value



# Internal Imports
from sanctify.examples.column_mapping_schema import EXAMPLE_COLUMN_MAPPING
from sanctify.examples.data_type_schema import EXAMPLE_COLUMN_MAPPING_USING_DATA_TYPE, EXAMPLE_DATA_TYPE_SCHEMA

__all__ = [
    "EXAMPLE_COLUMN_MAPPING",
    "EXAMPLE_COLUMN_MAPPING_USING_DATA_TYPE",
    "EXAMPLE_DATA_TYPE_SCHEMA",
]



# Internal Imports
from sanctify.constants import ComparisonOperations, Constants, DateOrderTuples
from sanctify.transformer import Transformer
from sanctify.validator import Validator

EXAMPLE_DATA_TYPE_SCHEMA = {
    "ACCOUNT": {
        Constants.TRANSFORMATIONS.value: [
            Transformer.remove_punctuations,
        ],
    },
    "NAME": {
        Constants.TRANSFORMATIONS.value: [
            Transformer.convert_to_lowercase,
            Transformer.replace_ii_with_II,
            Transformer.convert_jr_to_Junior,
            Transformer.convert_sr_to_Senior,
            Transformer.remove_punctuations,
            Transformer.remove_all_digits,
        ],
    },
    "DOB": {
        Constants.VALIDATIONS.value: [
            (
                Validator.validate_age,
                {
                    Constants.DATE_ORDER_TUPLE.value: DateOrderTuples.YEAR_MONTH_DAY.value,
                    Constants.COMPARISON_OPERATIONS.value: {ComparisonOperations.GREATER_THAN_EQUALS.value: 18},
                },
            )
        ],
    },
    "DUE_DATE": {
        Constants.VALIDATIONS.value: [
            (
                Validator.validate_due_date,
                {
                    Constants.DATE_ORDER_TUPLE.value: DateOrderTuples.YEAR_MONTH_DAY.value,
                    "comparison_operations": {"<=": 20, ">": 1},
                },
            )
        ],
    },
    "DATE": {
        Constants.TRANSFORMATIONS.value: [
            (
                Transformer.parse_date_from_string,
                {Constants.DATE_ORDER_TUPLE.value: DateOrderTuples.YEAR_MONTH_DAY.value},
            )
        ],
    },
    "PHONE": {
        Constants.TRANSFORMATIONS.value: [Transformer.extract_phone_number],
    },
    "AMOUNT": {
        Constants.TRANSFORMATIONS.value: [
            Transformer.remove_currency_from_amount,
        ],
    },
    "SSN": {Constants.TRANSFORMATIONS.value: [Transformer.remove_punctuations]},
    "ZIP_CODE": {Constants.VALIDATIONS.value: [Validator.validate_us_zip_code]},
    "STATE": {Constants.TRANSFORMATIONS.value: [Transformer.remove_punctuations]},
}

# AND use the Data Types Defined above in your column mapping
# The below dictionary represents the mapping of the Input Column Vs the Standard Column in your system with its Data type
EXAMPLE_COLUMN_MAPPING_USING_DATA_TYPE = {  # NOTE: Make sure that this  doesn't change during processing
    "DOB": {
        Constants.STANDARD_COLUMN.value: "Date of Birth",
        Constants.DATA_TYPE.value: "DOB",
    },
    "Cell Phone": {"standard_column": "Dial Number", "data_type": "PHONE"},
    "Zip": {Constants.STANDARD_COLUMN.value: "Zip Code", "data_type": "ZIP_CODE"},
    "State": {Constants.STANDARD_COLUMN.value: "State", "data_type": "STATE"},
    "SSN#": {Constants.STANDARD_COLUMN.value: "SSN", "data_type": "SSN"},
    "Account Due Date": {
        Constants.STANDARD_COLUMN.value: "Due Date",
        Constants.DATA_TYPE.value: "DUE_DATE",
    },
    "Due Amount": {
        Constants.STANDARD_COLUMN.value: "Amount",
        Constants.DATA_TYPE.value: "AMOUNT",
    },
    "Customer Number": {
        Constants.STANDARD_COLUMN.value: "Account",
        Constants.DATA_TYPE.value: "ACCOUNT",
    },
}



# Internal Imports
from sanctify.cleanser import Cleanser
from sanctify.constants import (
    PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP,
    AbstractDataTypes,
    CalendarDateComponents,
    ComparisonOperations,
    Constants,
    DateOrderTuples,
    DefaultColumns,
    PrimitiveDataTypes,
)
from sanctify.examples import EXAMPLE_COLUMN_MAPPING, EXAMPLE_COLUMN_MAPPING_USING_DATA_TYPE, EXAMPLE_DATA_TYPE_SCHEMA
from sanctify.exception import DataTypeParseError, ValidationError
from sanctify.processor import process_cleansed_df
from sanctify.serializer import CustomJSONDecoder, CustomJSONEncoder, SchemaDeSerializer, SchemaSerializer
from sanctify.transformer import Transformer
from sanctify.validator import Validator

__all__ = [
    "Cleanser",
    "PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP",
    "AbstractDataTypes",
    "CalendarDateComponents",
    "ComparisonOperations",
    "Constants",
    "CustomJSONDecoder",
    "CustomJSONEncoder",
    "DateOrderTuples",
    "DefaultColumns",
    "PrimitiveDataTypes",
    "DataTypeParseError",
    "ValidationError",
    "process_cleansed_df",
    "SchemaDeSerializer",
    "SchemaSerializer",
    "Transformer",
    "Validator",
    "EXAMPLE_COLUMN_MAPPING",
    "EXAMPLE_COLUMN_MAPPING_USING_DATA_TYPE",
    "EXAMPLE_DATA_TYPE_SCHEMA",
]



# External Imports
from typing import Iterable

import pandas as pd
from frozendict import frozendict
import demoji

# Internal Imports
from sanctify.constants import Constants, DefaultColumns, PrimitiveDataTypes
from sanctify.utils import convert_stringified_list_to_list, replace_dict_keys


class Cleanser:
    """
    The Cleanser class is responsible for validating and checking if a column is present and should be used.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        column_mapping_schema: frozendict | dict,
        data_type_schema: frozendict | dict = None,
    ) -> None:
        """
        Initializes the Cleanser object.

        Args:
            df (pd.DataFrame): The input DataFrame to be cleansed.
            column_mapping_schema (frozendict): The mapping schema for column names.
        """
        self.df = df
        self.mandatory_column_headers = []
        self.column_mapping_schema: frozendict[str:dict] = column_mapping_schema
        self.data_type_schema = data_type_schema
        self.input_vs_standard_column_mapping = self.extract_input_vs_standard_column_mapping()

    def check_if_all_columns_mapping_exist_in_df(self) -> None | ValueError:
        """
        Checks if all columns specified in the column mapping schema exist in the DataFrame.

        Returns:
            None: If all columns are present.

        Raises:
            ValueError: If any mandatory column is missing.
        """
        self.mandatory_column_headers = self.input_vs_standard_column_mapping.keys()

        missing_columns = set(self.mandatory_column_headers) - set(self.df.columns.to_list())
        if len(missing_columns) == 0:
            return None
        else:
            raise ValueError(f"Missing mandatory columns: ({', '.join(missing_columns)})")

    def extract_input_vs_standard_column_mapping(self) -> frozendict:
        """
        Extracts the input vs standard column mapping from the column mapping schema.

        Returns:
            frozendict: The input vs standard column mapping.
        """
        input_vs_standard_column_mapping = {}
        for input_column_name, column_config in self.column_mapping_schema.items():
            input_vs_standard_column_mapping[input_column_name] = column_config.get(
                Constants.STANDARD_COLUMN.value, input_column_name
            )
        return frozendict(input_vs_standard_column_mapping)

    def remove_trailing_spaces_from_column_headers(self) -> pd.DataFrame:
        """
        Removes trailing spaces from column headers in the DataFrame.

        Returns:
            pd.DataFrame: The DataFrame with trailing spaces removed from column headers.
        """
        self.df.columns = self.df.columns.str.strip()
        return self.df

    def drop_unmapped_columns(self) -> pd.DataFrame:
        """
        Drops columns from the DataFrame that are not present in the mandatory column headers.

        Returns:
            pd.DataFrame: The DataFrame with unmapped columns dropped.
        """
        self.mandatory_column_headers = self.input_vs_standard_column_mapping.keys()
        columns_to_keep = self.mandatory_column_headers
        self.df = self.df[columns_to_keep]
        return self.df

    def drop_fully_empty_rows(self, inplace: bool = True) -> pd.DataFrame:
        """
        Drops rows from the DataFrame that are fully empty.

        Args:
            inplace (bool): Whether to drop rows in-place or return a new DataFrame.

        Returns:
            pd.DataFrame: The DataFrame with fully empty rows dropped.
        """
        if inplace is True:
            self.df.dropna(how="all", inplace=inplace)
        else:
            self.df = self.df.dropna(how="all", inplace=inplace)

        return self.df

    def drop_rows_with_errors(self, inplace: bool = True, ignore_columns_list: Iterable = []) -> pd.DataFrame:
        """
        Drops all rows where the 'Error' column is not an empty string.

        Args:
            inplace (bool): If True, modify the DataFrame in-place. If False, return a new DataFrame.
            ignore_columns_list (list, tuple, set): If provided, only rows where the 'Error' column contains any of the other column names will be dropped

        Returns:
            pd.DataFrame: The cleaned DataFrame with rows dropped where the 'Error' column is not an empty string.
        """
        # Check if 'Error' column exists, and if not, create it with empty strings
        if DefaultColumns.ERROR.value not in self.df.columns:
            self.df[DefaultColumns.ERROR.value] = ""

        if isinstance(ignore_columns_list, Iterable) and len(ignore_columns_list) > 0:
            # Get list of column names to consider erroring out except from the ignore_columns_list and the default columns list
            columns_to_consider_erroring_out = (
                set(self.df.columns.to_list())
                - set(ignore_columns_list)
                - set([member.value for member in DefaultColumns])
            )

            # Create a mask for rows where the 'Error' column contains any of the strings in columns_to_consider_erroring_out
            mask = (
                self.df[DefaultColumns.ERROR.value].fillna("").str.contains("|".join(columns_to_consider_erroring_out))
            )

        else:
            # No columns to ignore, use an empty mask
            mask = self.df[DefaultColumns.ERROR.value].fillna("") != ""

        if inplace:
            # Modify the DataFrame in-place
            self.df.drop(self.df[mask].index, inplace=True)
            self.df.reset_index(drop=True, inplace=True)
        else:
            # Create a new DataFrame with rows excluded based on the mask
            df = self.df[~mask].copy()
            df.reset_index(drop=True, inplace=True)
            return df

        return self.df

    def remove_trailing_spaces_from_each_cell_value(self) -> pd.DataFrame:
        """
        Removes trailing spaces from each cell value in the DataFrame.

        Returns:
            pd.DataFrame: The DataFrame with trailing spaces removed from each cell value.
        """
        self.df = self.df.map(lambda x: x.strip() if isinstance(x, str) else x)
        return self.df

    def replace_column_headers(self, inplace: bool = True) -> tuple[pd.DataFrame, frozendict]:
        """
        Replaces column headers in the DataFrame based on the input vs standard column mapping.

        Args:
            inplace (bool): Whether to replace column headers in-place or return a new DataFrame.

        Returns:
            tuple[pd.DataFrame, frozendict]: The DataFrame with updated column headers
                and the updated column mapping schema.
        """
        if inplace is True:
            self.df.rename(columns=self.input_vs_standard_column_mapping, inplace=inplace)
        else:
            self.df = self.df.rename(columns=self.input_vs_standard_column_mapping, inplace=inplace)

        self.df.rename(columns=self.input_vs_standard_column_mapping, inplace=True)
        self.mandatory_column_headers = self.input_vs_standard_column_mapping.values()
        updated_column_mapping_schema = frozendict(
            replace_dict_keys(
                original_dict=self.column_mapping_schema,
                key_mapping=self.input_vs_standard_column_mapping,
            )
        )
        self.column_mapping_schema = updated_column_mapping_schema
        return self.df, updated_column_mapping_schema

    def mark_all_duplicates(self, columns: list[str], drop_duplicates: bool = False) -> pd.DataFrame:
        """
        INPLACE: Marks all duplicates based on the specified columns and populates the 'Error' column.

        Args:
            columns (list[str]): The list of column names to check for duplicates.
            drop_duplicates (bool): Whether to drop the duplicate rows or keep them. Default is False.

        Returns:
            pd.DataFrame: The DataFrame with duplicates marked and the 'Error' column populated.
        """
        if DefaultColumns.ERROR.value not in self.df.columns:
            self.df[DefaultColumns.ERROR.value] = ""

        if drop_duplicates is True:
            self.df.drop_duplicates(subset=columns, keep="first", inplace=True, ignore_index=True)
        else:
            duplicate_rows = self.df.duplicated(subset=columns, keep="first")

            error_message = f"Duplicate entry for {', '.join(columns)} passed"
            self.df.loc[duplicate_rows, DefaultColumns.ERROR.value] += " | " + error_message

        return self.df

    def drop_fully_duplicate_rows(self, inplace: bool = True) -> pd.DataFrame:
        """
        Drops rows from the DataFrame that are fully duplicates.

        Args:
            inplace (bool): Whether to drop rows in-place or return a new DataFrame.

        Returns:
            pd.DataFrame: The DataFrame with fully duplicate rows dropped.
        """
        if inplace is True:
            self.df.drop_duplicates(inplace=inplace, ignore_index=True)
        else:
            self.df = self.df.drop_duplicates(inplace=inplace, ignore_index=True)

        return self.df

    def modify_error_column_to_set_all_except_mandatory_to_blank(
        self, df: pd.DataFrame = None, use_standard_column_names: bool = True
    ) -> pd.DataFrame:
        """
        Modifies the 'Error' column in a DataFrame by setting all values to empty strings, except for specific columns that are considered for erroring out.

        Args:
            df (pd.DataFrame, optional): The DataFrame to modify the 'Error' column. If not provided, the DataFrame stored in the `Cleanser` object will be used.
            use_standard_column_names (bool, optional): A boolean flag indicating whether to use standard column names or raw column names from the column mapping schema.

        Returns:
            pd.DataFrame: The modified DataFrame with the 'Error' column updated.
        """
        if df is None:
            df = self.df

        ignore_columns_list = self.get_optional_column_names_from_column_mapping(
            return_standard_column_names=use_standard_column_names
        )

        # Check if 'Error' column exists, and if not, create it with empty strings
        if DefaultColumns.ERROR.value not in df.columns:
            return df

        if len(ignore_columns_list) > 0:
            # Get list of column names to consider erroring out except from the ignore_columns_list and the default columns list
            columns_to_consider_erroring_out = (
                set(df.columns.to_list()) - set(ignore_columns_list) - set([member.value for member in DefaultColumns])
            )

            # Create a mask for rows where the 'Error' column contains any of the strings in columns_to_consider_erroring_out
            mask = df[DefaultColumns.ERROR.value].fillna("").str.contains("|".join(columns_to_consider_erroring_out))

        else:
            # No columns to ignore, use an empty mask
            mask = df[DefaultColumns.ERROR.value].fillna("") != ""

        # Set all rows in the 'Error' column to empty strings except for those specified in columns_to_consider_erroring_out
        df.loc[~mask, DefaultColumns.ERROR.value] = ""

        return df

    def get_optional_column_names_from_column_mapping(
        self, return_standard_column_names: bool = True
    ) -> set | KeyError:
        """
        Retrieves the names of optional columns from the column mapping schema.

        Args:
            return_standard_column_names (bool, optional): A flag indicating whether to return the column names in their standard form or in their raw form. Defaults to True.

        Returns:
            set: A set of column names representing the optional columns from the column mapping schema. The column names can be in their standard form or in their raw form, depending on the value of the return_standard_column_names parameter.

        Example Usage:
            # Initialize the Cleanser object
            cleanser = Cleanser(df, column_mapping_schema)

            # Get the optional column names in their standard form
            optional_columns_standard = cleanser.get_optional_column_names_from_column_mapping(return_standard_column_names=True)

            # Get the optional column names in their raw form
            optional_columns_raw = cleanser.get_optional_column_names_from_column_mapping(return_standard_column_names=False)
        """
        optional_column_names = set()
        for raw_column_name, column_config in self.column_mapping_schema.items():
            if column_config.get(Constants.IS_OPTIONAL.value, False) is True:
                standard_column = column_config.get(Constants.STANDARD_COLUMN.value)
                if return_standard_column_names is True and standard_column is not None:
                    optional_column_names.add(column_config.get(Constants.STANDARD_COLUMN.value))
                else:
                    optional_column_names.add(raw_column_name)
        return optional_column_names

    def merge_data_type_schema_into_column_mapping_schema(self) -> frozendict:
        """
        Merges a data type schema into a column mapping schema.

        Updates the column configuration in the column mapping schema with the corresponding data type configuration from the data type schema.

        Returns:
            A frozendict representing the merged column mapping schema, where the keys are the raw column names and the values are the updated column configurations.
        """
        merged_column_mapping_schema: dict[str:dict] = dict(self.column_mapping_schema)
        for _, column_config in merged_column_mapping_schema.items():
            data_type = column_config.get(Constants.DATA_TYPE.value, {})
            column_config |= self.data_type_schema.get(data_type, {})
        return frozendict(merged_column_mapping_schema)

    def coalesce_rows(self, merge_on: str, agg_col_modifier: str = "Child ") -> None:
        """
        ### Modifies Dataframe in place!

        Merge rows in a DataFrame based on a specified column and aggregate the values of other columns.

        Args:
            merge_on (str): The column name to merge the rows on.
            agg_col_modifier (str, optional): A prefix to add to the new columns that will contain the aggregated values. Defaults to "Child ".

        Returns:
            None

        Example Usage:
            # Initialize the Cleanser object
            cleanser = Cleanser(df=df, column_mapping_schema=column_mapping_schema)

            # Call the coalesce_rows method
            cleanser.coalesce_rows(merge_on="merge_column", agg_col_modifier="Child ")

            # The DataFrame will be modified with merged rows and aggregated values in new columns
        """

        def agg_function(values):
            if len(values) > 1:
                return list(values)
            return ""

        aggregated_df = self.df.groupby(merge_on).first().reset_index()

        for col in self.df.columns:
            if col in self.mandatory_column_headers and (
                not self.column_mapping_schema[col].get(Constants.COALESCE_CHILD_REQUIRED, True)
            ):
                continue
            if col == merge_on:
                continue

            agg_col_name = f"{agg_col_modifier}{col}"
            aggregated_df[agg_col_name] = (
                self.df.groupby(merge_on)[col].agg(lambda x: agg_function(x)).reset_index()[col]
            )
            aggregated_df[agg_col_name] = aggregated_df[agg_col_name].astype(str, errors="ignore")

        self.df = aggregated_df
        self.df.reset_index()

    def explode_rows(self, agg_col_modifier: str = "Child ", remove_agg_cols: bool = False) -> pd.DataFrame:
        """
        Explodes rows in the DataFrame based on specified aggregation columns.

        Args:
            agg_col_modifier (str, optional): The prefix used to identify the aggregation columns. Default is "Child ".
            remove_agg_cols (bool, optional): Whether to remove the aggregation columns after exploding the rows. Default is False.

        Returns:
            pd.DataFrame: New DataFrame with exploded rows.

        Example Usage:
            cleanser = Cleanser(df, column_mapping_schema)
            exploded_df = cleanser.explode_rows(agg_col_modifier="Child ", remove_agg_cols=False)
        """
        agg_cols_map = {}
        for col in self.df.columns:
            if col.startswith(agg_col_modifier):  # Child Account
                agg_cols_map[col] = col.removeprefix(
                    agg_col_modifier
                )  # Behaviour: agg_cols_map["Child Account"] = "account"
                self.df[col] = self.df[col].apply(
                    convert_stringified_list_to_list
                )  # Behaviour: apply on "Child Account"

        self.df = self.df.explode(list(agg_cols_map.keys())).reset_index(drop=True)

        # Replace empty data in current col with non-empty data from subcol
        for subcol, col in agg_cols_map.items():
            non_empty_mask = self.df[subcol].apply(lambda x: bool(x) and any(x))
            self.df.loc[non_empty_mask, col] = self.df.loc[non_empty_mask, subcol]

        # Remove aggregated columns
        if remove_agg_cols is True:
            self.df = self.df.drop(columns=list(agg_cols_map.keys()), inplace=False)

        return self.df

    def merge_same_data_type_column(self, merge_column_data_type: str) -> pd.DataFrame:
        same_data_type_columns = [
            raw_column
            for raw_column, column_config in self.column_mapping_schema.items()
            if merge_column_data_type.lower() == column_config.get(Constants.DATA_TYPE.value).lower()
        ]

        for col in same_data_type_columns[1:]:
            self.df[same_data_type_columns[0]] = (
                self.df[same_data_type_columns[0]].replace("", pd.NA).combine_first(self.df[col])
            )
            self.column_mapping_schema[col][Constants.DATA_TYPE.value] = PrimitiveDataTypes.STRING.value
            self.column_mapping_schema[col].pop(Constants.STANDARD_COLUMN.value, None)
        self.input_vs_standard_column_mapping = self.extract_input_vs_standard_column_mapping()

        return self.df

    def remove_emojis(self) -> pd.DataFrame:
        self.df = self.df.map(lambda x: demoji.replace(x, "") if isinstance(x, str) else x)
        return self.df


# Internal Imports
from sanctify.constants import Constants, DateOrderTuples
from sanctify.transformer import Transformer
from sanctify.validator import Validator

EXAMPLE_COLUMN_MAPPING = {  # NOTE: Make sure that this  doesn't change during processing
    "DOB": {
        Constants.STANDARD_COLUMN.value: "Date of Birth",
        Constants.VALIDATIONS.value: [
            (
                Validator.validate_date_of_birth,
                {
                    "date_order_tuple": (
                        "year",
                        "month",
                        "day",
                    )  # Date Validations Require Date order tuples | See sanctify.constants.DateOrderTuples
                },
            )
        ],
    },
    "Cell Phone": {
        Constants.STANDARD_COLUMN.value: "Dial Number",
        Constants.VALIDATIONS.value: [
            (
                Validator.validate_phone_number_with_optional_country_code_check,
                {"country_code_equals": 1},
            )
        ],
    },
    "Zip": {
        Constants.STANDARD_COLUMN.value: "Zip Code",
        Constants.VALIDATIONS.value: [Validator.validate_us_zip_code],
    },
    "State": {
        Constants.STANDARD_COLUMN.value: "State",
        Constants.TRANSFORMATIONS.value: [Transformer.remove_punctuations],
    },
    "SSN#": {
        Constants.STANDARD_COLUMN.value: "SSN",
        Constants.VALIDATIONS.value: [Validator.validate_ssn],
    },
    "Account Due Date": {
        Constants.STANDARD_COLUMN.value: "Due Date",
        Constants.VALIDATIONS.value: [
            (
                Validator.validate_due_date,
                {Constants.DATE_ORDER_TUPLE.value: DateOrderTuples.YEAR_MONTH_DAY.value},  # Can be changed with strings
            )
        ],
    },
    "Due Amount": {
        Constants.STANDARD_COLUMN.value: "Amount",
        Constants.TRANSFORMATIONS.value: [
            Transformer.remove_currency_from_amount,
        ],
    },
    "Customer Number": {
        Constants.STANDARD_COLUMN.value: "Account",
        Constants.TRANSFORMATIONS.value: [
            Transformer.remove_punctuations,
        ],
    },
}



# External Imports
from typing import Any

from email_validator import EmailNotValidError, EmailUndeliverableError, ValidatedEmail, validate_email
from loguru import logger

# Internal Imports
from sanctify.constants import ComparisonOperations
from sanctify.exception import ValidationError
from sanctify.transformer import Transformer
from sanctify.utils import calculate_age, match_comparision_operation


class Validator:
    SSN_VALIDATION_ERROR_MSG = "Should be 4 digits"

    @staticmethod
    def validate_comparison_operations(value: Any, comparison_operations: dict[str: int | None]) -> Any | Exception:
        """
        Validate a value against a set of comparison operations.

        Args:
            value: The value to be validated.
            comparison_operations: A dictionary of comparison operations and their operands.

        Returns:
            The validated value if all comparison operations pass.

        Raises:
            ValidationError: If any comparison operation fails.
        """
        result = True
        for operation, operand in comparison_operations.items():
            result = result and match_comparision_operation(value=value, operation=operation, operand=operand)

        if result is True:
            return value
        else:
            raise ValidationError(f"Comparison Failed for {comparison_operations = }")

    @staticmethod
    def validate_age(
        value: str | int,
        date_order_tuple: tuple,
        comparison_operations: dict[str:int],
    ) -> str | Exception:
        """
        Validate the age based on comparison operations.

        Args:
            value: The value representing a date.
            date_order_tuple: A tuple specifying the order of year, month, and day in the date.
            comparison_operations: A dictionary of comparison operations and their operands.

        Returns:
            The validated date value if all comparison operations pass.

        Raises:
            ValidationError: If any comparison operation fails.
        """
        parsed_datetime = Transformer.parse_date_from_string(
            value=value, date_order_tuple=date_order_tuple, return_datetime=True
        )
        age = calculate_age(date_of_birth=parsed_datetime.date())

        if age < 0:
            raise ValidationError(f"Given date {str(parsed_datetime.date())} is in the future by {abs(age)} years")

        # Raises err on validation failure
        _ = Validator.validate_comparison_operations(value=age, comparison_operations=comparison_operations)

        return str(parsed_datetime.date())

    @staticmethod
    def validate_date_of_birth(
        value: str | int,
        date_order_tuple: tuple,
    ) -> str | Exception:
        """
        Validate the age based on comparison operations.

        Args:
            value: The value representing the date of birth.
            date_order_tuple: A tuple specifying the order of year, month, and day in the date.

        Returns:
            The validated date of birth date if it is valid

        Raises:
            ValidationError: If any comparison operation fails.
        """
        parsed_datetime = Transformer.parse_date_from_string(
            value=value, date_order_tuple=date_order_tuple, return_datetime=True
        )
        age = calculate_age(date_of_birth=parsed_datetime.date())

        if age < 0:
            raise ValidationError("Should not be in the future")

        try:
            # Raises err on validation failure
            _ = Validator.validate_comparison_operations(
                value=age,
                comparison_operations={
                    ComparisonOperations.GREATER_THAN_EQUALS.value: 18,
                    ComparisonOperations.LESS_THAN_EQUALS.value: 100,
                },
            )
        except ValidationError:
            raise ValidationError("Age should be >=18 and <=100")

        return str(parsed_datetime.date())

    @staticmethod
    def validate_due_date(
        value: str | int,
        date_order_tuple: tuple,
        comparison_operations: dict[str: int | None] = None,
    ) -> str | Exception:
        """
        Validate the due date based on comparison operations.

        Args:
            value: The value representing a date.
            date_order_tuple: A tuple specifying the order of year, month, and day in the date.
            comparison_operations: A dictionary of comparison operations and their operands.

        Returns:
            The validated due date if it is valid

        Raises:
            ValidationError: If any comparison operation fails.
        """
        parsed_datetime = Transformer.parse_date_from_string(
            value=value, date_order_tuple=date_order_tuple, return_datetime=True
        )
        age = calculate_age(date_of_birth=parsed_datetime.date())

        if age > 0:
            return str(parsed_datetime.date())
        else:
            age = abs(age)
            if isinstance(comparison_operations, dict):
                # Raises err on validation failure
                _ = Validator.validate_comparison_operations(value=age, comparison_operations=comparison_operations)
            else:
                try:
                    # Raises err on validation failure
                    _ = Validator.validate_comparison_operations(
                        value=age,
                        comparison_operations={
                            ComparisonOperations.LESS_THAN_EQUALS.value: 10,
                        },
                    )

                except ValidationError:
                    raise ValidationError("Should not be greater than 10 years")

            return str(parsed_datetime.date())

    @staticmethod
    def validate_email(
        value: str, check_deliverability: bool = False
    ) -> str | EmailNotValidError | EmailUndeliverableError | Exception:
        """
        Validate an email address.

        Args:
            value: The email address to be validated.
            check_deliverability: Boolean flag indicating whether to check email deliverability.

        Returns:
            The normalized and validated email address.

        Raises:
            ValidationError: If the email is not valid or undeliverable.
            Exception: If an unexpected exception occurs.
        """
        try:
            validated_email: ValidatedEmail = validate_email(email=value, check_deliverability=check_deliverability)
            return validated_email.normalized

        except (EmailNotValidError, EmailUndeliverableError) as err:
            # The email is not valid or undeliverable, handle the exception
            logger.exception(f"{str(err) = }")

            raise ValidationError(f"Invalid Email: {str(err)}")

    @staticmethod
    def validate_length_of_string(value: str, length: int):
        """
        Validate the length of a string.

        Args:
            value: The string to be validated.
            length: The expected length of the string.

        Returns:
            The validated string length.

        Raises:
            ValidationError: If the string length is not equal to the expected length.
        """
        return Validator.validate_comparison_operations(
            value=len(str(value)),
            comparison_operations={ComparisonOperations.EQUALS.value: length},
        )

    @staticmethod
    def validate_ssn(value: str):
        """
        Validate a Social Security Number (SSN).

        Args:
            value: The SSN to be validated.

        Returns:
            The validated SSN.

        Raises:
            ValidationError: If the SSN is not valid.
        """
        # Extract the ssn
        digits_only = Transformer.extract_digits_from_string(value=value)

        # Validate based on the number of digits
        if len(digits_only) < 4:
            raise ValidationError(Validator.SSN_VALIDATION_ERROR_MSG)
        else:
            # Take last 4
            return digits_only[-4:]

    @staticmethod
    def validate_zip_code(value: str):
        """
        **DEPRECATION NOTICE**
        **It is recommended to use the validate_us_zip_code to validate a United States Zip Code**
        **This method will be removed / overridden in the future in case of a different region based requirement**

        Validate a zip code.

        Args:
            value: The zip code to be validated.

        Returns:
            The validated zip code.

        Raises:
            ValidationError: If the zip code does not have 5 digits.
        """
        # Extract the zip code
        digits_only = Transformer.extract_digits_from_string(value=value)

        # Validate based on the number of digits
        if len(digits_only) < 5:
            raise ValidationError(f"ZipCode({digits_only}) does not have 5 digits (Actual: {len(digits_only)})")
        else:
            return digits_only[:5]

    @staticmethod
    def validate_us_zip_code(value: str):
        """
        Validates a United States zip code.

        Args:
            value: The US zip code to be validated.

        Returns:
            The validated US zip code.

        Raises:
            ValidationError: If the US zip code does not have 5 digits.
        """
        # Extract the zip code
        digits_only = Transformer.extract_digits_from_string(value=value)

        # Validate based on the number of digits
        if len(digits_only) < 5:
            raise ValidationError(f"US ZipCode({digits_only}) does not have 5 digits (Actual: {len(digits_only)})")
        else:
            return digits_only[:5]

    @staticmethod
    def validate_country_code(value: str, comparison_operations: dict, region: str | None = "US"):
        """
        Validate a country code in a given phone number.

        Args:
            value: The phone number whose country code to be validated.
            comparison_operations: eg. = 91 or < 10
            region (str | default = "US"): Here are some examples of valid country codes you can use as the region parameter:
                'US': United States
                'GB': United Kingdom
                'CA': Canada
                'AU': Australia
                'JP': Japan
                'DE': Germany
                You can find a comprehensive list of country codes in the ISO 3166-1 standard documentation.

        Returns:
            The validated country code integer.

        Raises:
            ValidationError: If phone number is invalid or comparison_operations fails
        """
        country_code = Transformer.extract_country_code(value=value, region=region)
        return Validator.validate_comparison_operations(value=country_code, comparison_operations=comparison_operations)

    @staticmethod
    def validate_phone_number(value: str, region: str | None = "US") -> str | ValidationError:
        """
        Validate a phone number.

        Args:
            value (str): The phone number to be validated.
            region (str | default = "US"): Here are some examples of valid country codes you can use as the region parameter:
                'US': United States
                'GB': United Kingdom
                'CA': Canada
                'AU': Australia
                'JP': Japan
                'DE': Germany
                You can find a comprehensive list of country codes in the ISO 3166-1 standard documentation.

        Returns:
            str | ValidationError: The validated phone number if it is valid.

        Raises:
            ValidationError: If the phone number is not valid.
        """
        return Transformer.extract_phone_number(value=value, region=region)

    @staticmethod
    def validate_phone_number_with_optional_country_code_check(
        value: str,
        region: str | None = "US",
        country_code_equals: int | str | None = None,
    ) -> str | ValidationError:
        """
        Validate a phone number with an optional country code check.

        Args:
            value (str): The phone number to be validated.
            region (str | default = "US"): Here are some examples of valid country codes you can use as the region parameter:
                'US': United States
                'GB': United Kingdom
                'CA': Canada
                'AU': Australia
                'JP': Japan
                'DE': Germany
                You can find a comprehensive list of country codes in the ISO 3166-1 standard documentation.
            country_code_equals (int | str | None, optional): The expected country code. Defaults to None.

        Returns:
            str | ValidationError: The validated phone number if the country code check passes.

        Raises:
            ValidationError: If the country code check fails.

        """
        parsed_number = Transformer.transform_str_into_phone_number_obj(value=value, region=region)

        if country_code_equals is not None and str(parsed_number.country_code) != str(country_code_equals):
            raise ValidationError(f"Country Code for {value} do not match {country_code_equals}")

        return str(parsed_number.national_number)



# External Imports
from typing import Callable

import pandas as pd
from frozendict import frozendict
from loguru import logger

# Internal Imports
from sanctify.constants import (
    PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP,
    AbstractDataTypes,
    Constants,
    DefaultColumns,
    PrimitiveDataTypes,
)
from sanctify.transformer import Transformer
from sanctify.validator import Validator

PRIMITIVE_DATA_TYPE_VS_TRANSFORMATIONS = {
    PrimitiveDataTypes.BOOLEAN.value: [Transformer.parse_boolean],
    PrimitiveDataTypes.FLOAT.value: [float],
    AbstractDataTypes.DECIMAL.value: [float],
    PrimitiveDataTypes.INTEGER.value: [int],
    PrimitiveDataTypes.STRING.value: [str],
    AbstractDataTypes.TEXT.value: [str],
}


def handle_additional_transformations_and_validations(
    data_type_name_for_column: str,
) -> list[Callable]:
    """
    Handles additional transformations and validations based on the data type name for a column.

    Args:
        data_type_name_for_column (str): Data type name for the column.

    Returns:
        list[Callable]: List of transformation functions to apply.
        []: Empty List if data type do not match
    """
    return PRIMITIVE_DATA_TYPE_VS_TRANSFORMATIONS.get(data_type_name_for_column, [])


def handle_data_type(data_type_schema: dict[str:dict], col_config: dict) -> tuple[list[Callable], list[Callable]]:
    """
    Handles data type transformations and validations for a column.

    Args:
        data_type_schema (dict[str: dict]): Data type schema containing data type configurations.
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.DATA_TYPE.value in col_config:
        data_type_name_for_column: str = col_config[Constants.DATA_TYPE.value]
        transformations += handle_additional_transformations_and_validations(
            data_type_name_for_column=data_type_name_for_column
        )
        data_type_config_for_column: dict = data_type_schema.get(data_type_name_for_column, {})
        transformations += data_type_config_for_column.get(Constants.TRANSFORMATIONS, [])
        validations += data_type_config_for_column.get(Constants.VALIDATIONS, [])

    return transformations, validations


def process_nullable_or_post_processing_dtype(series: pd.Series, dtype) -> pd.Series:
    """
    Process a pandas Series by inferring the object data type, filling missing values with 0,
    and converting the data type to the specified dtype.

    Args:
        series (pd.Series): The pandas Series to be processed.
        dtype: The desired data type for the Series.

    Returns:
        pd.Series: The processed pandas Series with the specified data type.
    """
    processed_series = series.infer_objects().fillna(0).astype(dtype, errors="ignore")
    return processed_series


def handle_date_order_tuple(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles date order transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.DATE_ORDER_TUPLE.value in col_config:
        date_order_tuple = col_config[Constants.DATE_ORDER_TUPLE.value]
        transformations = [
            (
                Transformer.parse_date_from_string,
                {Constants.DATE_ORDER_TUPLE.value: date_order_tuple},
            )
        ]

    return transformations, validations


def handle_comparison_operation(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles comparison operation validations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.COMPARISON_OPERATIONS.value in col_config:
        comparison_operations = col_config[Constants.COMPARISON_OPERATIONS.value]
        validations = [
            (
                Validator.validate_comparison_operations,
                {"comparison_operations": comparison_operations},
            )
        ]

    return transformations, validations


def handle_transformations(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles general transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.TRANSFORMATIONS.value in col_config:
        transformations = col_config[Constants.TRANSFORMATIONS.value]

    return transformations, validations


def handle_validations(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles validations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.VALIDATIONS.value in col_config:
        validations = col_config[Constants.VALIDATIONS.value]

    return transformations, validations


def handle_post_processing_dtype(col_config: dict, data_type_schema: dict) -> Callable or None:
    """
    Handles post-validation transformations for a column.

    Args:
        col_config (dict): Column configuration containing column-specific settings.
        data_type_schema (dict): A schema containing data types and their configurations.

    Returns:
        Callable or None: If post-validation transformation is applicable for int or float,
        it returns the corresponding transformation function. Otherwise, it returns None.
    """

    result = None

    # Check if the column configuration contains a post-processing data type
    if Constants.POST_PROCESSING_DATA_TYPE.value in col_config:
        pp_datatype = col_config.get(Constants.POST_PROCESSING_DATA_TYPE.value)
        # Get the corresponding post-validation transformation function from the mapping
        result = PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP.get(pp_datatype, None)

    # Check if the column configuration refers to a predefined data type in the schema
    if Constants.DATA_TYPE.value in col_config:
        data_type_name_for_column: str = col_config.get(Constants.DATA_TYPE.value, None)
        if data_type_name_for_column is not None:
            # Retrieve the configuration for the specified data type from the schema
            data_type_config_for_column: dict = data_type_schema.get(data_type_name_for_column, {})
            # Check if the data type configuration contains a post-processing data type
            if Constants.POST_PROCESSING_DATA_TYPE.value in data_type_config_for_column:
                pp_datatype = data_type_config_for_column.get(Constants.POST_PROCESSING_DATA_TYPE.value)
                # Get the corresponding post-validation transformation function from the mapping
                result = PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP.get(pp_datatype, None)

    return result


def apply_functions(cell_value: str, col_name: str, functions: list[Callable | tuple[Callable, dict]]) -> pd.Series:
    """
    Applies a series of transformation and validation functions to a value.

    Args:
        value: The value to apply the functions to.
        functions (list[Callable]): List of transformation and validation functions.
        functions (tuple[Callable, dict]]): List of tuple of transformation and validation functions and additional kwargs.

    Returns:
        pd.Series: Series containing the modified value and error message (if any).
    """
    final_cell_value = cell_value
    error_message = None

    if not cell_value or cell_value == "nan":  # All possible absent value cases
        error_message = f"{col_name} is required"

    else:
        for function in functions:
            func = function
            kwargs = {}

            if isinstance(function, (list, tuple)):
                func = function[0]
                kwargs = function[1]

            logger.debug(f"Running {func.__name__ = } on {col_name} containing {cell_value = } with {kwargs = }")

            try:
                final_cell_value = func(final_cell_value, **kwargs)

            except Exception as err:
                logger.error(f"FAILED: {func.__name__ = } on {cell_value = } with {kwargs = } | ERROR: {(str(err)) = }")
                error_message = f"{col_name} with {cell_value = } got error = {str(err)}"
                # Reset the cell_value on failure
                final_cell_value = cell_value
                break

            if final_cell_value in {"", "nan"}:
                str_err = "Cleansed value is empty"
                logger.error(f"FAILED: {func.__name__ = } on {cell_value = } with {kwargs = } | ERROR: {str_err}")
                error_message = f"{col_name} with {cell_value = } got error = {str_err}"
                # Reset the cell_value on failure
                final_cell_value = cell_value
                break

            logger.debug(
                f"Result: {func.__name__ = } on {col_name} containing {cell_value = } with {kwargs = } | {final_cell_value = }"
            )

    # Return the modified value and error message (if any)
    return pd.Series(
        [final_cell_value, error_message],
        index=[DefaultColumns.VALUE_PLACEHOLDER.value, DefaultColumns.ERROR.value],
    )


def process_cleansed_df(
    df: pd.DataFrame,
    column_mapping_schema: frozendict = {},
    data_type_schema: frozendict = {},
    ignore_optional_columns: bool = False,
) -> pd.DataFrame:
    """
    Processes the cleansed DataFrame by applying transformations and validations based on the column mapping and data type schema.

    Args:
        df (pd.DataFrame): The cleansed DataFrame.
        column_mapping_schema (frozendict): Column mapping schema.
        data_type_schema (frozendict): Data type schema.
        ignore_optional_columns (bool): Flag indicating whether to ignore optional columns when applying transformations and validations.

    Returns:
        pd.DataFrame: The processed DataFrame.
    """
    if column_mapping_schema == {}:
        return df

    # Create an "Error" column in the DataFrame to track errors, if it does not exists
    if DefaultColumns.ERROR.value not in df.columns:
        df[DefaultColumns.ERROR.value] = ""

    # Iterate over each column in column_mapping_schema
    for col_name, col_config in column_mapping_schema.items():
        logger.debug(f"{col_name = } | {col_config = }")

        # If a column is marked optional and ignore_optional_columns is True,
        # then ignore any transformation or validation on that column
        is_optional_column = col_config.get(Constants.IS_OPTIONAL.value, False)
        if ignore_optional_columns is True and is_optional_column is True:
            continue

        # Check if col_name in column mapping schema exists in df
        col_name_in_df_columns = col_name in df.columns
        if col_name_in_df_columns is False:
            logger.error(f"Column({col_name}) with doesn't exist in the input dataframe | {col_config = }")
            # Continue to next column in the column mapping schema
            continue

        if col_config == {}:
            logger.error(f"Column({col_name})'s column config not provided in the input dataframe | {col_config = }")
            # Continue to next column in the column mapping schema
            continue  # Skip empty config items
        else:
            logger.debug(f"Parsing Column({col_name}) as str dtype | {col_config = }")
            # Convert column rows to string data type
            df[col_name] = df[col_name].astype(str)

        transformations, validations = [], []

        d_type_transformations, d_type_validations = handle_data_type(data_type_schema, col_config)
        d_order_transformations, d_order_validations = handle_date_order_tuple(col_config)
        (
            comparison_op_transformations,
            comparison_op_validations,
        ) = handle_comparison_operation(col_config)
        col_transformations, _ = handle_transformations(col_config)
        _, col_validations = handle_validations(col_config)
        post_processing_dtype = handle_post_processing_dtype(col_config=col_config, data_type_schema=data_type_schema)

        transformations += (
            d_type_transformations + d_order_transformations + comparison_op_transformations + col_transformations
        )
        validations += d_type_validations + d_order_validations + comparison_op_validations + col_validations

        # Apply transformations and validations on the column
        logger.debug(f"Running {col_name}: {transformations = }")
        logger.debug(f"Running {col_name}: {validations = }")

        data_type_name_for_column: str = col_config.get(Constants.DATA_TYPE.value)
        primitive_data_type_classes = handle_additional_transformations_and_validations(
            data_type_name_for_column=data_type_name_for_column
        )
        applied_series: pd.Series = df[col_name].apply(
            apply_functions, col_name=col_name, functions=transformations + validations
        )

        # Update cell values
        successful_rows = applied_series[DefaultColumns.ERROR.value].isnull()
        df.loc[successful_rows, col_name] = applied_series.loc[successful_rows, DefaultColumns.VALUE_PLACEHOLDER.value]

        # Update "Error" column for errored rows
        errored_rows = applied_series[DefaultColumns.ERROR.value].notnull()
        if errored_rows.sum() > 0:
            df.loc[errored_rows, col_name] = applied_series.loc[errored_rows, DefaultColumns.VALUE_PLACEHOLDER.value]
            df.loc[errored_rows, DefaultColumns.ERROR.value] += (
                " | " + applied_series.loc[errored_rows, DefaultColumns.ERROR.value]
            )

        if len(primitive_data_type_classes) == 1:
            primitive_data_type_class = primitive_data_type_classes[0]
            if primitive_data_type_class in PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP:
                nullable_dtype = PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP[primitive_data_type_class]
                df[col_name] = process_nullable_or_post_processing_dtype(series=df[col_name], dtype=nullable_dtype)
                logger.debug(f"Converting the column {col_name} back to {nullable_dtype} data type")

        if post_processing_dtype is not None:
            logger.debug(f"Converting the column {col_name} to {post_processing_dtype} data type")
            df[col_name] = process_nullable_or_post_processing_dtype(series=df[col_name], dtype=post_processing_dtype)

    return df
